/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#include <stdio.h>
#include "ap_int.h"
#include "hls_interface.h"
#include "hls_utils.h"
#include "hls_aes.h"

ap_uint<8> xtime2(ap_uint<8> a)
{
#pragma HLS INLINE
  if ((a >> 7) == 1)
    return ((a << 1) ^ 0x1b);
  else
    return (a << 1);
}
ap_uint<8> xtime3(ap_uint<8> a)
{
#pragma HLS INLINE
  return (xtime2(a)^a);
}
ap_uint<8> xtime4(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime2(xtime2(a));
}
ap_uint<8> xtime8(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime4(xtime2(a));
}
ap_uint<8> xtime9(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime8(a)^a;
}
ap_uint<8> xtimeb(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime8(a)^xtime2(a)^a;
}
ap_uint<8> xtimed(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime8(a)^xtime4(a)^a;
}
ap_uint<8> xtimee(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime8(a)^xtime4(a)^xtime2(a);
}

void ShiftRows(array_t ostate, array_t istate)
{
#pragma HLS INLINE
  ostate[0]  = istate[0];
  ostate[4]  = istate[4];
  ostate[8]  = istate[8];
  ostate[12] = istate[12];

  ostate[13] = istate[1];
  ostate[1]  = istate[5];
  ostate[5]  = istate[9];
  ostate[9]  = istate[13];

  ostate[10] = istate[2];
  ostate[2]  = istate[10];
  ostate[14] = istate[6];
  ostate[6]  = istate[14];

  ostate[7]  = istate[3];
  ostate[3]  = istate[15];
  ostate[15] = istate[11];
  ostate[11] = istate[7];
}

void iShiftRows(array_t ostate, array_t istate)
{
#pragma HLS INLINE
  ostate[0]  = istate[0];
  ostate[4]  = istate[4];
  ostate[8]  = istate[8];
  ostate[12] = istate[12];

  ostate[13] = istate[9];
  ostate[1]  = istate[13];
  ostate[5]  = istate[1];
  ostate[9]  = istate[5];

  ostate[10] = istate[2];
  ostate[2]  = istate[10];
  ostate[14] = istate[6];
  ostate[6]  = istate[14];

  ostate[3]  = istate[7];
  ostate[7]  = istate[11];
  ostate[11] = istate[15];
  ostate[15] = istate[3];
}

void SubBytes(array_t ostate, array_t istate)
{
#pragma HLS INLINE
#pragma HLS RESOURCE variable=Sbox00 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox01 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox02 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox03 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox04 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox05 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox06 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox07 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox08 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox09 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox10 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox11 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox12 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox13 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox14 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox15 core=ROM_1P_1S

  ostate[0]  = Sbox00[istate[0]];
  ostate[1]  = Sbox01[istate[1]];
  ostate[2]  = Sbox02[istate[2]];
  ostate[3]  = Sbox03[istate[3]];
  ostate[4]  = Sbox04[istate[4]];
  ostate[5]  = Sbox05[istate[5]];
  ostate[6]  = Sbox06[istate[6]];
  ostate[7]  = Sbox07[istate[7]];
  ostate[8]  = Sbox08[istate[8]];
  ostate[9]  = Sbox09[istate[9]];
  ostate[10] = Sbox10[istate[10]];
  ostate[11] = Sbox11[istate[11]];
  ostate[12] = Sbox12[istate[12]];
  ostate[13] = Sbox13[istate[13]];
  ostate[14] = Sbox14[istate[14]];
  ostate[15] = Sbox15[istate[15]];
}

void iSubBytes(array_t ostate, array_t istate)
{
#pragma HLS INLINE
#pragma HLS RESOURCE variable=iSbox00 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox01 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox02 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox03 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox04 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox05 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox06 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox07 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox08 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox09 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox10 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox11 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox12 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox13 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox14 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox15 core=ROM_1P_1S

  ostate[0]  = iSbox00[istate[0]];
  ostate[1]  = iSbox01[istate[1]];
  ostate[2]  = iSbox02[istate[2]];
  ostate[3]  = iSbox03[istate[3]];
  ostate[4]  = iSbox04[istate[4]];
  ostate[5]  = iSbox05[istate[5]];
  ostate[6]  = iSbox06[istate[6]];
  ostate[7]  = iSbox07[istate[7]];
  ostate[8]  = iSbox08[istate[8]];
  ostate[9]  = iSbox09[istate[9]];
  ostate[10] = iSbox10[istate[10]];
  ostate[11] = iSbox11[istate[11]];
  ostate[12] = iSbox12[istate[12]];
  ostate[13] = iSbox13[istate[13]];
  ostate[14] = iSbox14[istate[14]];
  ostate[15] = iSbox15[istate[15]];
}

void MixColumns(array_t ostate, array_t istate)
{
#pragma HLS INLINE
    ostate[0]  = xtime2(istate[ 0]) ^ xtime3(istate[ 1]) ^        istate[ 2]  ^        istate[ 3] ;
    ostate[1]  =        istate[ 0]  ^ xtime2(istate[ 1]) ^ xtime3(istate[ 2]) ^        istate[ 3] ;
    ostate[2]  =        istate[ 0]  ^        istate[ 1]  ^ xtime2(istate[ 2]) ^ xtime3(istate[ 3]);
    ostate[3]  = xtime3(istate[ 0]) ^        istate[ 1]  ^        istate[ 2]  ^ xtime2(istate[ 3]);

    ostate[4]  = xtime2(istate[ 4]) ^ xtime3(istate[ 5]) ^        istate[ 6]  ^        istate[ 7] ;
    ostate[5]  =        istate[ 4]  ^ xtime2(istate[ 5]) ^ xtime3(istate[ 6]) ^        istate[ 7] ;
    ostate[6]  =        istate[ 4]  ^        istate[ 5]  ^ xtime2(istate[ 6]) ^ xtime3(istate[ 7]);
    ostate[7]  = xtime3(istate[ 4]) ^        istate[ 5]  ^        istate[ 6]  ^ xtime2(istate[ 7]);

    ostate[8]  = xtime2(istate[ 8]) ^ xtime3(istate[ 9]) ^        istate[10]  ^        istate[11] ;
    ostate[9]  =        istate[ 8]  ^ xtime2(istate[ 9]) ^ xtime3(istate[10]) ^        istate[11] ;
    ostate[10] =        istate[ 8]  ^        istate[ 9]  ^ xtime2(istate[10]) ^ xtime3(istate[11]);
    ostate[11] = xtime3(istate[ 8]) ^        istate[ 9]  ^        istate[10]  ^ xtime2(istate[11]);

    ostate[12] = xtime2(istate[12]) ^ xtime3(istate[13]) ^        istate[14]  ^        istate[15] ;
    ostate[13] =        istate[12]  ^ xtime2(istate[13]) ^ xtime3(istate[14]) ^        istate[15] ;
    ostate[14] =        istate[12]  ^        istate[13]  ^ xtime2(istate[14]) ^ xtime3(istate[15]);
    ostate[15] = xtime3(istate[12]) ^        istate[13]  ^        istate[14]  ^ xtime2(istate[15]);
}

void iMixColumns(array_t ostate, array_t istate)
{
#pragma HLS INLINE
    ostate[0]  = xtimee(istate[ 0]) ^ xtimeb(istate[ 1]) ^ xtimed(istate[ 2]) ^ xtime9(istate[ 3]);
    ostate[1]  = xtime9(istate[ 0]) ^ xtimee(istate[ 1]) ^ xtimeb(istate[ 2]) ^ xtimed(istate[ 3]);
    ostate[2]  = xtimed(istate[ 0]) ^ xtime9(istate[ 1]) ^ xtimee(istate[ 2]) ^ xtimeb(istate[ 3]);
    ostate[3]  = xtimeb(istate[ 0]) ^ xtimed(istate[ 1]) ^ xtime9(istate[ 2]) ^ xtimee(istate[ 3]);

    ostate[4]  = xtimee(istate[ 4]) ^ xtimeb(istate[ 5]) ^ xtimed(istate[ 6]) ^ xtime9(istate[ 7]);
    ostate[5]  = xtime9(istate[ 4]) ^ xtimee(istate[ 5]) ^ xtimeb(istate[ 6]) ^ xtimed(istate[ 7]);
    ostate[6]  = xtimed(istate[ 4]) ^ xtime9(istate[ 5]) ^ xtimee(istate[ 6]) ^ xtimeb(istate[ 7]);
    ostate[7]  = xtimeb(istate[ 4]) ^ xtimed(istate[ 5]) ^ xtime9(istate[ 6]) ^ xtimee(istate[ 7]);

    ostate[8]  = xtimee(istate[ 8]) ^ xtimeb(istate[ 9]) ^ xtimed(istate[10]) ^ xtime9(istate[11]);
    ostate[9]  = xtime9(istate[ 8]) ^ xtimee(istate[ 9]) ^ xtimeb(istate[10]) ^ xtimed(istate[11]);
    ostate[10] = xtimed(istate[ 8]) ^ xtime9(istate[ 9]) ^ xtimee(istate[10]) ^ xtimeb(istate[11]);
    ostate[11] = xtimeb(istate[ 8]) ^ xtimed(istate[ 9]) ^ xtime9(istate[10]) ^ xtimee(istate[11]);

    ostate[12] = xtimee(istate[12]) ^ xtimeb(istate[13]) ^ xtimed(istate[14]) ^ xtime9(istate[15]);
    ostate[13] = xtime9(istate[12]) ^ xtimee(istate[13]) ^ xtimeb(istate[14]) ^ xtimed(istate[15]);
    ostate[14] = xtimed(istate[12]) ^ xtime9(istate[13]) ^ xtimee(istate[14]) ^ xtimeb(istate[15]);
    ostate[15] = xtimeb(istate[12]) ^ xtimed(istate[13]) ^ xtime9(istate[14]) ^ xtimee(istate[15]);
}

void AddRoundKey(array_t ostate, array_t istate, array_t rk)
{
#pragma HLS INLINE
    xor2(ostate, istate, rk);
}

void AESRound(array_t ostate, array_t ostate_fdb, array_t istate, array_t rk)
{
    array_t sub, shift, mix, add;
#pragma HLS ARRAY_RESHAPE variable=istate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk                 complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate_fdb         complete dim=1
#pragma HLS ARRAY_RESHAPE variable=sub                complete dim=1
#pragma HLS ARRAY_RESHAPE variable=shift              complete dim=1
#pragma HLS ARRAY_RESHAPE variable=mix                complete dim=1
#pragma HLS ARRAY_RESHAPE variable=add                complete dim=1


    /* Note: Add a different sbox lookup implementation here due
     * to a bug in the c-simulator provided by Xilinx
     * (Found in Version: 2016.2 Linux)
     */
#if __SYNTHESIS__
    SubBytes(sub, istate);
#else
    static const ap_uint<8> Sbox[256] = {
      0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76, 0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
      0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15, 0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
      0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84, 0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
      0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8, 0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
      0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73, 0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
      0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79, 0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
      0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a, 0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
      0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf, 0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16
    };
    for (size_t i=0; i<16; i++) {
        sub[i] = Sbox[istate[i]];
    }
#endif
    ShiftRows(shift, sub);
    MixColumns(mix, shift);
    AddRoundKey(add, mix, rk);

    copy(ostate_fdb, add);
    xor2(ostate, shift, rk);
//    std::cout << "\tinput : " << reshape_array(istate) << std::endl;
//    std::cout << "\tsub   : " << reshape_array(sub) << std::endl;
//    std::cout << "\tshift : " << reshape_array(shift) << std::endl;
//    std::cout << "\tmix   : " << reshape_array(mix) << std::endl;
//    std::cout << "\tadd   : " << reshape_array(add) << std::endl;
}

void AESRoundBasic(array_t ostate, array_t istate, array_t rk)
{
    array_t sub, shift, mix;
#pragma HLS ARRAY_RESHAPE variable=istate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk                 complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=sub                complete dim=1
#pragma HLS ARRAY_RESHAPE variable=shift              complete dim=1
#pragma HLS ARRAY_RESHAPE variable=mix                complete dim=1

    SubBytes(sub, istate);
    ShiftRows(shift, sub);
    MixColumns(mix, shift);
    AddRoundKey(ostate, mix, rk);
}

void AESInvRound(array_t ostate, array_t ostate_fdb, array_t istate, array_t rk)
{
    array_t isub, ishift, imix, add;
#pragma HLS ARRAY_RESHAPE variable=istate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk                 complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate_fdb         complete dim=1
#pragma HLS ARRAY_RESHAPE variable=isub               complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ishift             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=imix               complete dim=1
#pragma HLS ARRAY_RESHAPE variable=add                complete dim=1

    iShiftRows(ishift, istate);
    iSubBytes(isub, ishift);
    AddRoundKey(add, isub, rk);
    iMixColumns(imix, add);
  
    copy(ostate_fdb, imix);
    copy(ostate, add);
}

void AESInvRoundBasic(array_t ostate, array_t istate, array_t rk)
{
    array_t isub, ishift, imix;
#pragma HLS ARRAY_RESHAPE variable=istate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk                 complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=isub               complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ishift             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=imix               complete dim=1
    
    iMixColumns(imix, istate);
    iShiftRows(ishift, imix);
    iSubBytes(isub, ishift);
    AddRoundKey(ostate, isub, rk);
}

void AESKeyUpdate(array_t new_rk, array_t rk, ap_uint<4> round)
{
#pragma HLS INLINE
    int i, j;
    ap_uint<8> tk[16];

    static const ap_uint<8> rcon[10] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36};

#pragma HLS RESOURCE variable=SboxK0 core=ROM_1P_1S
#pragma HLS RESOURCE variable=SboxK1 core=ROM_1P_1S
#pragma HLS RESOURCE variable=SboxK2 core=ROM_1P_1S
#pragma HLS RESOURCE variable=SboxK3 core=ROM_1P_1S
#pragma HLS RESOURCE variable=rcon   core=ROM_1P_1S
#pragma HLS ARRAY_RESHAPE variable=tk complete dim=1 reshape

    for(j = 0; j < 4; j++)
    {
#pragma HLS UNROLL
      for(i = 0; i < 4; i++)
      {
#pragma HLS UNROLL
        tk[i+j*4] = rk[i+j*4];
      }
    }

    /* while not enough round key material calculated */
    /* calculate new values */
    tk[0] ^= SboxK0[tk[1+4*3]] ^ rcon[round];
    tk[1] ^= SboxK1[tk[2+4*3]];
    tk[2] ^= SboxK2[tk[3+4*3]];
    tk[3] ^= SboxK3[tk[0+4*3]];

    for(j = 1; j < 4; j++)
#pragma HLS UNROLL
        for(i = 0; i < 4; i++)
#pragma HLS UNROLL
          tk[i+4*j] ^= tk[i+4*(j-1)];

    /* copy values into round key array */
    for(j = 0; j < 4; j++)
#pragma HLS UNROLL
      for(i = 0; i < 4; i++)
#pragma HLS UNROLL
        new_rk[i+4*j] = tk[i+4*j];
}
